part of 'application.dart';

/// The ViewModel class is designed to store and manage data in a lifecycle conscious way.
/// ViewModel is a class that is responsible for preparing and managing the data for a Widget tree.
/// It also handles the communication of the Widgets with the rest of the application (e.g. calling the
/// business logic classes).
/// A ViewModel is always created in association with a scope (defined with the use of the related [ViewModelScope]
/// widget) and will be retained as long as the scope is alive (until unmount).
/// The purpose of the ViewModel is to acquire and keep the information that is necessary for your wdgets.
/// This widgets should be able to observe changes in the ViewModel.
/// ViewModels usually expose this information via LiveData.
/// ViewModel's only responsibility is to manage the data for the UI.
/// It should never access your view hierarchy or hold a reference back to the widgets tree.
class ViewModel {
  @protected
  void onCleared() {}
}

/// Factory used to instantiate your ViewModel.
/// The best practice is defining the ViewModel constructor as private and exposing a static parameterless method that
/// returns the newly constructed instance of your ViewModel. You should pass only that method as [ViewModelFactory]
/// parameter
typedef ViewModelFactory<T extends ViewModel> = T Function();

/// This method allows you to acquire ViewModel instance from global scope. This means that the ViewModel is retained
/// as long as your application is closed or restarted
VM globalViewModel<VM extends ViewModel>(ViewModelFactory<VM> factory) =>
    _GlobalViewModelProvider.instance.provide(factory);

extension StateExt on State {
  /// Get scoped ViewModel from current state context
  /// See [ViewModelScope]
  VM viewModel<VM extends ViewModel>(ViewModelFactory<VM> factory, {Symbol? scope}) =>
      context.viewModel<VM>(factory, scope: scope);
}

extension BuildContextExt on BuildContext {
  /// Get scoped ViewModel from current context
  /// See [ViewModelScope]
  VM viewModel<VM extends ViewModel>(ViewModelFactory<VM> factory, {Symbol? scope}) =>
      _ViewModelScopeElement._provide<VM>(
          getElementForInheritedWidgetOfExactType<ViewModelScope>() as _ViewModelScopeElement?, factory, scope);
}

/// The provider class instantiates and stores your ViewModel instances
mixin ViewModelProvider {
  final Map<Type, ViewModel> _viewModels = {};

  /// Provide a [VM] instance.
  /// The first time you call this method on [VM], [factory] will be called to create the ViewModel instance which
  /// will be retained for future calls until [clearViewModels] is called
  VM provide<VM extends ViewModel>(ViewModelFactory<VM> factory) => (_viewModels[VM] ??= factory()) as VM;

  /// Clears all ViewModels provided by this provided until now
  void clearViewModels() {
    for (final key in _viewModels.keys.toList(growable: false)) _viewModels.remove(key)?.onCleared();
  }
}

/// Define a new ViewModel scope.
/// All descendants of this widget will have access to the [ViewModelProvider] of this scope through it's context
/// Use cases of scoped viewModels
///     globalViewModel(MyViewModel.factory) // scope: <global>
///
///     context.viewModel(MyViewModel.factory) // scope: <global>
///
///     ViewModelScope.builder(
///       scope: #my_scope,
///       builder: (context) => context.viewModel(MyViewModel.factory)); // scope: #my_scope
///
///     ViewModelScope.builder(
///       scope: #my_scope,
///       builder: (context) => ViewModelScope.builder(
///         scope: #my_scope_2,
///         builder: (context) => context.viewModel(MyViewModel.factory))); // scope: #my_scope_2
///
///     ViewModelScope.builder(
///       scope: #my_scope,
///       builder: (context) => ViewModelScope.builder(
///         scope: #my_scope_2,
///         builder: (context) => context.viewModel(MyViewModel.factory, scope: #my_scope))); // scope: #my_scope
class ViewModelScope extends InheritedWidget {
  @override
  ObjectKey get key => super.key as ObjectKey;

  Symbol get scope => key.value as Symbol;

  ViewModelScope({required Symbol scope, required Widget child}) : super(key: ObjectKey(scope), child: child);

  ViewModelScope.builder({required Symbol scope, required WidgetBuilder builder})
      : this(scope: scope, child: Builder(builder: builder));

  @override
  bool updateShouldNotify(ViewModelScope oldWidget) => oldWidget.scope != scope;

  @override
  InheritedElement createElement() => _ViewModelScopeElement(this);

  @override
  void debugFillProperties(DiagnosticPropertiesBuilder properties) {
    super.debugFillProperties(properties);
    properties.add(DiagnosticsProperty<Symbol>('scope', scope));
  }
}

class _ViewModelScopeElement extends InheritedElement with ViewModelProvider {
  @override
  ViewModelScope get widget => super.widget as ViewModelScope;

  _ViewModelScopeElement? _parent;

  _ViewModelScopeElement(ViewModelScope widget) : super(widget);

  @override
  void mount(Element? parent, Object? newSlot) {
    super.mount(parent, newSlot);
    _parent = parent?.getElementForInheritedWidgetOfExactType<ViewModelScope>() as _ViewModelScopeElement?;
  }

  @override
  void unmount() {
    super.unmount();
    clearViewModels();
  }

  static VM _provide<VM extends ViewModel>(
      _ViewModelScopeElement? element, ViewModelFactory<VM> factory, Symbol? scope) {
    if (element == null) if (scope == null)
      return _GlobalViewModelProvider.instance.provide(factory);
    else
      throw StateError('no scope #${scope.name} associated with current context');
    if (scope == null || element.widget.scope == scope) return element.provide(factory);
    return _provide<VM>(element._parent, factory, scope);
  }
}

class _GlobalViewModelProvider with ViewModelProvider {
  static final _GlobalViewModelProvider instance = _GlobalViewModelProvider._();

  _GlobalViewModelProvider._();
}
