part of 'app_route.dart';

// ignore: prefer_mixin
class AppRouterDelegate extends RouterDelegate<AppPage> with ChangeNotifier, PopNavigatorRouterDelegateMixin {
  static final List<AppPage> _stackPlaceholder = [
    ActivityRoute(location: '/_', builder: (context, args) => const SizedBox.shrink()).createPage()
  ];
  late AppRouteInformationProvider provider;
  late AppRouteInformationParser parser;

  final ApplicationState _app;
  final Map<String, ActivityRoute> _routes;
  final String _initialLocation;
  final OnUnknownRoute _onUnknownRoute;
  final List<AppPage> _stack = [];

  @override
  final GlobalKey<NavigatorState> navigatorKey = GlobalKey();

  @override
  AppPage? get currentConfiguration => _stack.lastOrNull;

  RouteInformation get _entrypoint =>
      RouteInformation(location: WidgetsBinding.instance?.window.defaultRouteName ?? Navigator.defaultRouteName);

  AppRouterDelegate(this._app, Iterable<ActivityRoute> routes, this._initialLocation, this._onUnknownRoute)
      : _routes = Map.fromEntries(routes.map((route) => MapEntry(route.location, route))) {
    provider = AppRouteInformationProvider._(this);
    parser = AppRouteInformationParser._(this);
  }

  @override
  Future<void> setInitialRoutePath(AppPage configuration) {
    final stack = kIsWeb ? _parseRouteStack() : null;
    if (stack == null) return super.setInitialRoutePath(configuration);
    for (final location in stack) {
      final page = _parseRouteInformationInternal(RouteInformation(location: location));
      if (page != null) _addRoute(page);
    }
    return SynchronousFuture(null);
  }

  @override
  Future<void> setNewRoutePath(AppPage configuration) {
    if (configuration != currentConfiguration) _addRoute(configuration);
    return SynchronousFuture(null);
  }

  @override
  Widget build(BuildContext context) => Navigator(
      key: navigatorKey,
      pages: _stack.isEmpty ? _stackPlaceholder : _stack.toList(growable: false),
      onPopPage: _onPopPage,
      restorationScopeId: 'app_nav');

  @override
  void notifyListeners() {
    super.notifyListeners();
    final title = currentConfiguration?.name;
    if (title != null) _app.title = title;
  }

  Future<bool> didPushRoute(String location) {
    final current = currentConfiguration?.location;
    if (location == current) return SynchronousFuture(true);
    final index = _stack.indexWhere((page) => page.location == location) + 1;
    if (index > 0) {
      while (_stack.length > index) _stack.removeLast();
      notifyListeners();
      return SynchronousFuture(true);
    } else {
      return SynchronousFuture(false);
    }
  }

  Future<T> push<T>(ActivityRoute<T> route, {RouteArgs args = const {}}) async {
    final page = route.createPage(args: args);
    _addRoute(page);
    return page.result;
  }

  bool _onPopPage(Route route, dynamic result) {
    final appPage = route.settings;
    if (route.didPop(result) && appPage is AppPage) {
      _removeRoute(appPage);
      return true;
    }
    return false;
  }

  void _addRoute(AppPage page) {
    _stack.add(page);
    notifyListeners();
  }

  void _removeRoute(AppPage page) {
    _stack.remove(page);
    notifyListeners();
  }

  Future<AppPage> _parseRouteInformation(RouteInformation routeInformation) {
    final page = _parseRouteInformationInternal(routeInformation);
    return SynchronousFuture(page == null ? _unknownRoute(routeInformation) : page);
  }

  AppPage _unknownRoute(RouteInformation routeInformation) {
    final route = _onUnknownRoute(routeInformation);
    var page = route?.createPage(args: {'location': routeInformation.location});
    if (page != null) {
      page.location = routeInformation.location;
    } else {
      page = currentConfiguration ??
          _parseRouteInformationInternal(const RouteInformation(location: Navigator.defaultRouteName));
      if (page == null) throw StateError('Cannot find route ${routeInformation.location}');
      if (page.location != routeInformation.location) updateUrl(page.location);
    }
    return page;
  }

  AppPage? _parseRouteInformationInternal(RouteInformation routeInformation) {
    var location = routeInformation.location ?? Navigator.defaultRouteName;
    if (location == Navigator.defaultRouteName) location = _initialLocation;
    final route = _routes[location];
    if (route == null) return null;
    final args = kIsWeb ? _parseRouteInformationState(location) : routeInformation.state as RouteArgs?;
    return route.createPage(args: args ?? const {});
  }

  RouteInformation _restoreRouteInformation(AppPage configuration) {
    final location = configuration.location;
    if (kIsWeb) {
      _restoreRouteInformationState(location, configuration.arguments);
      _restoreRouteStack(_stack.map((page) => page.route.location).toList(growable: false));
    }
    return RouteInformation(location: location, state: configuration.arguments);
  }

  static AppRouterDelegate of(BuildContext context) => Router.of(context).routerDelegate as AppRouterDelegate;

  RouteArgs? _parseRouteInformationState(String location) {
    final data = sessionStorage[location];
    return data != null ? jsonDecode(data) : null;
  }

  void _restoreRouteInformationState(String location, RouteArgs? args) {
    if (args == null || args.isEmpty) {
      sessionStorage.remove(location);
    } else {
      sessionStorage[location] = jsonEncode(args);
    }
  }

  List<String>? _parseRouteStack() {
    final data = sessionStorage['_stack_'];
    return data != null ? (jsonDecode(data) as List).cast<String>() : null;
  }

  void _restoreRouteStack(List<String>? locations) {
    if (locations == null || locations.isEmpty) {
      sessionStorage.remove('_stack_');
    } else {
      sessionStorage['_stack_'] = jsonEncode(locations);
    }
  }
}
